# -*- coding:utf-8 -*-

import chardet
from pyftpdlib.filesystems import AbstractedFS, FilesystemError
from uploader import Uploader
from os import path
from utils import reformat_lm, get_file_list, get_object_meta
from client import Client
from baidubce import exception
from baidubce.services import bos
from bosftp_config import BosftpCfg
from appender import Appender
import urllib
import logging

logger = logging.getLogger(__name__)
class UrlHandler(object):
    def __init__(self, url):
        self.closed = False
        self.name = url 
        try:
            self.fd = urllib.urlopen(url)
        except Exception as e:
            raise e
    
    def read(self, size=None):
        return self.fd.read(size) 

    def close(self):
        pass

class BosAbstractFileSystem(AbstractedFS):
    def __init__(self, *args, **kwargs):
        #super(BosAbstractFileSystem, self).__init__(*args, **kwargs)
        AbstractedFS.__init__(self, *args, **kwargs)
        self.bucket_name = BosftpCfg().bucket

    @property
    def client(self):
        return Client().bos_client

    def realpath(self, path):
        return path

    # return fle handler
    def open(self, filename, mode):
        assert isinstance(filename, unicode), filename
        logger.info("open filename: {0}, mode: {1}, current working dir: {2}".format(str(filename).encode("utf-8"), str(mode).encode("utf-8"), str(self.cwd).encode("utf-8"))) 
        if self.isdir(filename):
            raise FilesystemError(filename + "is a directory")

        ftp_path = self.fs2ftp(filename)
        logger.debug("ftp_path: {0}".format(str(ftp_path).encode("utf-8")))
        object_key = ftp_path[1:]  # remove "/" from head
        logger.debug("object_key: {0}".format(str(object_key)))
        if mode.startswith('r') or mode.startswith('R'):
            exist = get_object_meta(object_key)
            if not exist:
                raise FilesystemError("Not exist such file")
            try:
                url = Client().bos_client.generate_pre_signed_url(self.bucket_name, str(object_key))
                logger.debug("open generate_pre_signed_url:{0}".format(str(url).encode("utf-8")))
                fd = UrlHandler(url)
            except Exception as e:
                logger.error("open occurs an error, err_msg: {0}".format(str(e.message).encode("utf-8")))
                raise FilesystemError("open file:{0} failed".format(str(path).encode("utf-8")))
            return fd
        elif mode.startswith('a') or mode.startswith('A'):
            return Appender(self.bucket_name, object_key)
        else:
            return Uploader(self.bucket_name, object_key)

    def getsize(self, path):
        logger.info("file:{0} getsize, cwd: {1}".format(str(path).encode("utf-8"), str(self.cwd).encode("utf-8")))
        assert isinstance(path, unicode), path

        if self.isfile(path):
            object_key = self.fs2ftp(path).strip("/")
            try:
                response = Client().bos_client.get_object_meta_data(self.bucket_name, str(object_key))
            except exception.BceError as e:
                logger.exception(
                    "get_object_meta_data in getsize occurs an exception.bucket_name: {0}, object_key: {1}, err_msg: {2}".format(str(self.bucket_name).encode("utf-8"), str(self.object_key).encode("utf-8"), e.message))
                raise e
            return response.metadata.content_length 
        elif self.isdir(path):
            return 0
        else:
            logger.error("Path: {0} is invalid!".format(str(path).encode("utf-8")))
            raise FilesystemError("Path: {0} is invalid!".format(str(path).encode("utf-8")))

    def chdir(self, path):
        logger.info("chdir to path:{0}, cwd: {1}".format(str(path).encode("utf-8"), str(self.cwd).encode("utf-8")))
        assert isinstance(path, unicode), path

        if self.isdir(path):
            self.cwd = self.fs2ftp(path)
            #print str(self.cwd)
            logger.debug("current work directory {0}".format(str(self.cwd).encode("utf-8")))
        else:
            raise FilesystemError, 'Not a dir.'

    def mkdir(self, path):
        logger.info("mkdir path:{0}, cwd: {1}".format(str(path).encode("utf-8"), str(self.cwd).encode("utf-8")))
        assert isinstance(path, unicode), path

        if self.isdir(path):
            logger.error(path + " directory exists")
            raise FilesystemError(path + " directory exists")
        elif self.isfile(path):
            logger.error(path + " is file")
            raise FilesystemError(path + " is file")

        ftp_path = self.fs2ftp(path)
        logger.debug("ftp_path: {0}".format(str(ftp_path).encode("utf-8")))
        dir_name = ftp_path.strip("/") + "/"
        logger.debug("key_name: {0}".format(str(dir_name).encode("utf-8")))

        try:
            response = Client().bos_client.put_object_from_string(self.bucket_name, str(dir_name), "")
        except exception.BceError as e:
            logger.exception("make dir: {0} occurs an exception.err_msg: {1}".format(str(dir_name).encode("utf-8"), e.message))
            raise FilesystemError("make dir:{0} failed".format(str(ftp_path).encode("utf-8")))

        logger.debug("content_length: {0}".format(str(response.content_length).encode("utf-8")))

    def rename(self, src, dest):
        logger.info("rename from {0} to {1}, cwd: {2}".format(str(src).encode("utf-8"), str(dest).encode("utf-8"), str(self.cwd).encode("utf-8")))
        assert isinstance(src, unicode), src 
        assert isinstance(dest, unicode), dest

        if src == dest:
            return

        if self.isfile(src):
            src_key = self.fs2ftp(src).strip("/")
            dest_key = self.fs2ftp(dest).strip("/")

            try:
                response = Client().bos_client.copy_object(self.bucket_name, str(src_key), 
                        self.bucket_name, str(dest_key))

                Client().bos_client.delete_object(
                    self.bucket_name,
                    src_key
                )
            except exception.BceError as e:
                logger.exception(
                    "Rename " + str(src).encode("utf-8") + "to" + str(dest).encode("utf-8") + "occurs an exception.")
                raise FilesystemError(
                    "Rename {0} to {1} failed.".format(str(src).encode("utf-8"), str(dest).encode("utf-8")))
        elif self.isdir(src):
            raise FilesystemError("Directory renaming is not supported")
        else:
            raise FilesystemError("Invalid parameter!")

    def listdir(self, path):
        logger.info("listdir path: {0}, cwd: {1}".format(str(path).encode("utf-8"), str(self.cwd).encode("utf-8")))
        assert isinstance(path, unicode), path

        ftp_path = self.fs2ftp(path)
        logger.debug("ftp_path: {0}".format(str(ftp_path).encode("utf-8")))
        dir_name = ftp_path
        logger.debug("dir_name: {0}".format(str(dir_name).encode("utf-8")))

        list_name = list()

        prefix = ""
        if dir_name.lstrip("/") != "":
            prefix = dir_name.strip("/") + "/"
            dir_name = prefix

        str(prefix)

        is_truncated = True
        next_marker = str("")
        while is_truncated and next_marker is not None:
            try:
                response = Client().bos_client.list_objects(self.bucket_name, 
                        prefix = str(prefix).encode("utf-8"), 
                        marker = str(next_marker).encode("utf-8"), 
                        delimiter = "/")
                file_list = get_file_list(response, str(prefix).encode("utf-8")) 
                list_name.extend(file_list)
                if response.is_truncated:
                    is_truncated = True
                    next_marker = response.next_marker
                else:
                    is_truncated = False
            except exception.BceError as e:
                logger.exception(
                        "List dir path: {0} occurs an exception. err_msg: {1}".format(str(dir_name).encode("utf-8"), e.message))
                raise FilesystemError("list dir:{0} failed.".format(ftp_path))
        return list_name

    def islink(self, fs_path):
        return False


    
    def isfile(self, path):
        logger.info("isfile path:{0}, cwd: {1}".format(str(path).encode("utf-8"), str(self.cwd).encode("utf-8")))
        assert isinstance(path, unicode), path

        if path.startswith("/"):
            ftp_path = self.fs2ftp(path)
            object_key = ftp_path.strip("/")
            if get_object_meta(object_key):
                return True
        return False

    def isdir(self, path):  
        logger.info("isdir path:{0}, cwd: {1}".format(str(path).encode("utf-8"), str(self.cwd).encode("utf-8")))
        assert isinstance(path, unicode), path
        #return not self.isfile(path)
        if path.startswith("/"):
            ftp_path = self.fs2ftp(path)
            if ftp_path == "/": # root 
                return True

            object_key = ftp_path.strip("/")
            str(object_key)
            if get_object_meta(object_key):
                # regular file
                return False
            elif get_object_meta(object_key + "/"):
                # directory file
                return True 
            else:
                # fake dir
                try:
                    response = Client().bos_client.list_objects(self.bucket_name, 
                                    max_keys = 2, 
                                    prefix = object_key + "/", delimiter = "/")
                except exception.BceClientError as e:
                    logger.exception(
                        "isdir:{0} occurs a BceClientError, message: {1}".format(str(path).encode("utf-8"), e.message))
                    return False
                except exception.BceError as e:
                    logger.exception(
                        "isdir:{0} occurs a BceError, message: {1}".format(str(path).encode("utf-8"), e.message))
                    raise e
                if len(response.contents) != 0 or len(response.common_prefixes) != 0:
                    return True
                return False
        else:
            return False

    def rmdir(self, path):
        logger.info("rmdir path:{0}, cwd: {1}".format(str(path).encode("utf-8"), str(self.cwd).encode("utf-8")))
        assert isinstance(path, unicode), path

        if self.isdir(path):
            dir_name = self.fs2ftp(path).strip("/") + "/"
            logger.debug("dir_name:{0}".format(str(dir_name).encode("utf-8")))
            try:
                response = Client().bos_client.delete_object(self.bucket_name, dir_name)
                logger.debug("response:{0}".format(str(response).encode("utf-8")))
            except Exception as e:
                logger.info(
                    "rmdir: {0} occurs exception: {1}".format(str(dir_name).encode("utf-8"), 
                            str(e.message).encode("utf-8")))
        else:
            raise FilesystemError("No such directory")
            

    def remove(self, path):
        logger.info(
            "remove path:{0}, cwd: {1}".format(str(path).encode("utf-8"), str(self.cwd).encode("utf-8")))
        assert isinstance(path, unicode), path

        if self.isfile(path):
            object_key = self.fs2ftp(path).strip("/")
            logger.debug("object_key: {0}".format(str(object_key).encode("utf-8")))
            response = Client().bos_client.delete_object(self.bucket_name, str(object_key))
            logger.debug("response: {0}".format(str(response).encode("utf-8")))
        else:
            raise FilesystemError("No such file")

    def lexists(self, path):
        logger.info("lexists path: {0}".format(str(path).encode("utf-8")))
        ftp_path = self.fs2ftp(path)

        if ftp_path.startswith("/"):
            if self.isdir(path) or self.isfile(path):
                return True
            else:
                return False
        return False

    def format_list(self, basedir, listing, ignore_err=True):
        for basename in listing:
            ft, size, last_modified, name = basename
            last_modified = reformat_lm(last_modified, form="ls")

            if ft == 'dir':
                perm = "rwxrwxrwx"
                t = 'd'
            else:
                perm = "r-xr-xr-x"
                t = '-'
            line = "%s%s\t1\towner\tgroup\t%s\t%s\t%s\r\n" % (t, perm, size, last_modified, name)
            if self.cmd_channel is not None:
                yield line.encode("utf8", self.cmd_channel.unicode_errors)
            else:
                yield line.encode("utf8")

    def format_mlsx(self, basedir, listing, perms, facts, ignore_err=True):
        for basename in listing:
            ft, size, last_modified, name = basename
            last_modified = reformat_lm(last_modified, form="mlsx")

            if ft == 'dir':
                perm = 'el'
            else:
                perm = 'r'
            line = "type=%s;size=%d;perm=%s;modify=%s %s\r\n" % (ft, size, perm, last_modified, name)
            if self.cmd_channel is not None:
                yield line.encode("utf8", self.cmd_channel.unicode_errors)
            else:
                yield line.encode("utf8")

if __name__ == "__main__":
    fs = BosAbstractFileSystem() 
    fs.listdir(u"/")
