# -*- coding:utf-8 -*-
from pyftpdlib.filesystems import AbstractedFS, FilesystemError
import datetime
from cStringIO import StringIO
from baidubce import exception
from baidubce.services import bos
from client import Client
import logging
from bosftp_config import BosftpCfg
logger = logging.getLogger(__name__)

class FifoBuffer(object):
    def __init__(self):
        self._cur_read_pos = 0
        self._cur_write_pos = 0
        self._buf = StringIO()
        self._buf_size = 0

    def size(self):
        return self._buf_size

    def read(self, read_len):
        self._buf.seek(self._cur_read_pos)                                          # 先定位文件指针到当前读指针的位置
        read_buf = self._buf.read(read_len)                                         # 读出内容
        last_content = self._buf.read(self._cur_write_pos - self._buf.tell())       # 读出尾部的东西
        self._buf.seek(0)
        self._buf.write(last_content)                                               # 将尾部拷贝到前面
        self._cur_write_pos = self._buf.tell()
        self._cur_read_pos = 0
        self._buf_size = self._buf_size - read_len
        return read_buf

    def write(self, data):
        self._buf.seek(self._cur_write_pos)
        self._buf.write(data)
        self._cur_write_pos = self._buf.tell()
        self._buf_size = self._buf_size + len(data)

    def close(self):
        self._buf.seek(0, 2)
        self._buf.close()
        del self._buf

def get_object_meta(object_key):
    try:
        response = Client().bos_client.get_object_meta_data(BosftpCfg().bucket, str(object_key))
    except exception.BceHttpClientError as e:
        logger.info("BceHttpClientError, msg : {0}".format(str(e.message).encode("utf-8")))
        logger.info("not exists such key:{0}".format(str(object_key).encode("utf-8")))
        return False
    except exception.BceError as e:
        logger.error("Exception: {0}".format(str(e.message).encode("utf-8")))
        raise FilesystemError(e.message)
    return True


def get_file_list(list_objects_response, cwd):
    """
    parse files from list_objects_response
    """
    list_name = list()
    list_dir = set()
    list_key = set()

    if list_objects_response.common_prefixes is not None:
        for dir_obj in list_objects_response.common_prefixes:
            dir_path = "/" + str(dir_obj.prefix)
            dir_name = dir_path[len(cwd):].strip("/")
            if dir_name != "":
                list_dir.add(("dir", 0, None, dir_name))

    for obj in list_objects_response.contents:
        key_path = "/" + str(obj.key)
        key_name = key_path[len(cwd):].strip("/")

        if key_name == "":
            continue

        if key_path.endswith("/"):
            list_dir.add(("dir", 0, None, key_name))
        else:
            list_key.add(("file", int(obj.size), obj.last_modified, key_name))

    for dir in list_dir:
        list_name.append(dir)

    for key in list_key:
        list_name.append(key)

    return list(list_dir.union(list_key))

def reformat_lm(last_modified, form="object"):
    if last_modified is None:
        dt_modified = datetime.datetime(1970, 1, 1)
    else:
        try:
            #this is the format used when you use get_all_keys()
            dt_modified = datetime.datetime.strptime(last_modified, '%Y-%m-%dT%H:%M:%SZ')
        except ValueError:
            #this is the format used when you use get_key()
            dt_modified = datetime.datetime.strptime(last_modified, '%a, %d %b %Y %H:%M:%S %Z')
        except:
            raise

    if form == "object":
        return dt_modified
    elif form == "ls":
        if (datetime.datetime.now() - dt_modified).days < 180:
            return dt_modified.strftime("%b %d %H:%M")
        else:
            return dt_modified.strftime("%b %d %Y")
    else:
        return dt_modified.strftime("%Y%m%d%H%M%S")
